/**
 * win32fast/monitor.cpp
 */
#include "i_monitor.h"

#if _WIN32_WINNT < 0x0400
#define SignalObjectAndWait(m_hEntry, m_hDelay, dwTimeout, bUnused) (SetEvent(m_hEntry), WaitForSingleObject(m_hDelay, dwTimeout))
#endif

namespace await_n {

enum { MAX_ABSOLUTE = 0x7fffffff };

MonitorImpl::MonitorImpl(): m_nLockCount(-1), m_nRecursionCount(0), m_nSleepCount(0), m_thrOwner(0), m_hEntry(0), m_hDelay(0) {
	m_hEntry = CreateEvent(NULL, FALSE, FALSE, NULL);
	m_hDelay = CreateEvent(NULL, TRUE, FALSE, NULL);
}

MonitorImpl::~MonitorImpl() {
	CloseHandle(m_hDelay);
	CloseHandle(m_hEntry);
}

void MonitorImpl::enter() {
	threadid_t thrCurrent = getSelfThreadId();
	if(thrCurrent == m_thrOwner) { // already owner
		++m_nRecursionCount;
	}
	else {
		if(0 != InterlockedIncrement(&m_nLockCount))
			WaitForSingleObject(m_hEntry, INFINITE); // enqueue because somebody has already locked
		m_nRecursionCount = 1;
		m_thrOwner = thrCurrent;
	}
}

void MonitorImpl::leave() {
	--m_nRecursionCount;
	if(0 == m_nRecursionCount) {
		m_thrOwner = 0;
		if(InterlockedDecrement(&m_nLockCount) >= 0)
			SetEvent(m_hEntry);
	}
}

bool MonitorImpl::sleep(AbsoluteMoment moment) {
	moment -= GetTickCount();
	DWORD dwTimeout = (moment >= MAX_ABSOLUTE)? INFINITE: (DWORD) moment;
	threadid_t thrCurrent = getSelfThreadId();

	LONG nRecursionCount = m_nRecursionCount;
	++m_nSleepCount;
	m_nRecursionCount = 0;
	m_thrOwner = 0;

	DWORD res = SignalObjectAndWait(m_hEntry, m_hDelay, dwTimeout, FALSE);
	WaitForSingleObject(m_hEntry, INFINITE);

	m_thrOwner = thrCurrent;
	m_nRecursionCount = nRecursionCount;

	return (res == WAIT_TIMEOUT);
}

void MonitorImpl::awake() {
	if(m_nSleepCount>0) {
		m_nSleepCount = 0;
		PulseEvent(m_hDelay);
	}
}

//----------------------------------------------------------------------------

Monitor *getGlobalMonitor() {
	static MonitorImpl g_oRawMonitor;
	return &g_oRawMonitor;
}

AbsoluteMoment cvtRelativeToAbsolute(long s) {
	AbsoluteMoment result = s;
	if(result<0)
		result = MAX_ABSOLUTE;
	else
		result += GetTickCount();
	return result;
}

threadid_t getSelfThreadId() {
	return GetCurrentThreadId();
}

} // namespace await_n

//=============================================================================
